import os
import sys
import warnings
import bson
from bson import BSON
from falcon.runtime import ONNXRuntime, FalconRuntime
from typing import List, Optional, Dict
import bson
from bson import BSON
from typing import List, Tuple, Optional
from numpy import typing as npt
import numpy as np
from typing import Any, Dict, Union


def run_model(model_path: str, X: npt.NDArray) -> Union[List[npt.NDArray], np.ndarray]:
    """
    Runs input data through the saved model.

    Parameters
    ----------
    model_path : str
        model path
    X : npt.NDArray
        model inputs

    Returns
    -------
    Union[List[npt.NDArray], np.ndarray]
        model predictions
    """
    if model_path.endswith("onnx"):
        return run_onnx(model_path, X, "final")
    else:
        return run_falcon(model_path, X)


def run_falcon(model: Union[bytes, str], X: npt.NDArray) -> Union[List[npt.NDArray], np.ndarray]:
    runtime = FalconRuntime(model=model)
    return runtime.run(X)


def run_onnx(
    model: Union[bytes, str], X: npt.NDArray, outputs: str = "final"
) -> List[npt.NDArray]:
    runtime = ONNXRuntime(model=model)
    return runtime.run(X, outputs=outputs)


def set_verbosity_level(level: int = 1) -> None:
    if level not in {0, 1}:
        level = 0
    os.environ["FALCON_VERBOSITY_LEVEL"] = str(level)


def print_(*args: Any) -> None:
    verbosity_level = os.getenv("FALCON_VERBOSITY_LEVEL", "1")
    if verbosity_level == "1":
        for a in args:
            print(a)


def disable_warnings() -> None:
    if not sys.warnoptions:
        warnings.simplefilter("ignore")
        os.environ["PYTHONWARNINGS"] = "ignore"



from abc import ABC, abstractmethod
from falcon.serialization import SerializedModelRepr


class ONNXConvertible(ABC):
    """
    Base class for all models/pipeline_elements/pipelines that can be converted to onnx.
    """
    @abstractmethod
    def to_onnx(self) -> SerializedModelRepr:
        """
        Converted model

        Returns
        -------
        SerializedModelRepr
        """
        pass



import pandas as pd
from typing import Union
import numpy as np
from falcon.utils import print_


def load_churn_dataset(mode: str = "training") -> Union[pd.DataFrame, np.ndarray]:
    print_("Loading churn dataset ...")
    df = pd.read_csv(
        "https://gist.githubusercontent.com/OKUA1/b5faf7b5b3fa9d69bbb64b52670ecf10/raw/d5f87274ad244f3da4b9e330bf7fc9a8d3015f0b/churn.csv"
    )
    if mode == "training":
        print_(df.head(5))
        print_(f"Dataset shape: {df.shape}")
        print_("This dataset can be used for `tabular_classification` task")
    elif mode == "inference":
        df.pop("churn")
        df = df.to_numpy()
    else:
        raise ValueError(f"Unknown mode {mode}, expected `training` or `inference`")
    return df


def load_insurance_dataset(mode: str = "training") -> Union[pd.DataFrame, np.ndarray]:
    print_("Loading insurance dataset ...")
    df = pd.read_csv(
        "https://gist.githubusercontent.com/OKUA1/b5faf7b5b3fa9d69bbb64b52670ecf10/raw/d5f87274ad244f3da4b9e330bf7fc9a8d3015f0b/insurance.csv"
    )
    if mode == "training":
        print_(df.head(5))
        print_(f"Dataset shape: {df.shape}")
        print_("This dataset can be used for `tabular_regression` task")
    elif mode == "inference":
        df.pop("charges")
        df = df.to_numpy()
        print(df)
    else:
        raise ValueError(f"Unknown mode {mode}, expected `training` or `inference`")
    return df




from falcon.abstract.learner import Learner
from falcon.abstract.model import Model
from falcon.abstract.processor import Processor
from falcon.abstract.task_manager import TaskManager
from falcon.abstract.task_pipeline import Pipeline, PipelineElement
from falcon.abstract.onnx_convertible import ONNXConvertible



from copy import deepcopy
import pandas as pd
from typing import Union, Tuple, Optional, List
import numpy as np
from numpy import isin, typing as npt
from falcon import types as ft
from ..abstract.task_pipeline import Pipeline
from sklearn.model_selection import RepeatedStratifiedKFold, RepeatedKFold
from sklearn.metrics import balanced_accuracy_score, r2_score


def read_data(path: str) -> pd.DataFrame:
    if path.endswith(".csv"):
        data = pd.read_csv(path)
    elif path.endswith(".parquet"):
        data = pd.read_parquet(path)
    else:
        raise ValueError("Only `.csv` and `.parquet` files are supported")

    return data


def clean_data(
    data: Union[pd.DataFrame, npt.NDArray]
) -> Union[pd.DataFrame, npt.NDArray]:
    if isinstance(data, pd.DataFrame):
        return data.dropna()
    else:
        mask = pd.isnull(data)
        keep = []
        for row in mask:
            if True in row:
                keep.append(False)
            else:
                keep.append(True)
        data = data[keep, :]
        return data


def clean_data_split(X: npt.NDArray, y: npt.NDArray) -> Tuple[npt.NDArray, npt.NDArray]:
    mask_x = pd.isnull(X)
    mask_y = pd.isnull(y)
    keep = []
    for i in range(len(mask_x)):
        if len(y.shape) == 1:
            if True not in mask_x[i] and mask_y[i] == False:
                keep.append(True)
            else:
                keep.append(False)
        else:
            if True not in mask_x[i] and True not in mask_y[i]:
                keep.append(True)
            else:
                keep.append(False)
    X = X[keep, :]
    if len(y.shape) == 1:
        y = y[keep]
    else:
        y = y[keep, :]
    return X, y


def convert_to_np_obj(
    data: Union[pd.DataFrame, npt.NDArray]
) -> npt.NDArray[np.object_]:
    if isinstance(data, pd.DataFrame):
        return data.to_numpy(dtype=np.object_)
    else:
        return data.astype(np.object_)


def split_features(
    data: Union[pd.DataFrame, npt.NDArray],
    features: Optional[ft.ColumnsList],
    target: Optional[Union[str, int]],
) -> Tuple[npt.NDArray[np.object_], npt.NDArray[np.object_]]:
    if features is not None and len(features) < 1:
        ValueError("Features List cannot be empty")
    if (
        isinstance(data, np.ndarray)
        and features is not None
        and isinstance(features[0], str)
    ):
        ValueError("Expected list of integers as features, found strings")
    if isinstance(data, np.ndarray) and isinstance(target, str):
        ValueError("Expected integer as target, found string")

    # TODO: provide a proper fix instead of a ValuError
    if (target is None or features is None) and not (
        target is None and features is None
    ):
        raise ValueError(
            "Either both target and features should be provided or neither of them."
        )
    if isinstance(data, pd.DataFrame):
        if features is None:
            X = data.iloc[:, :-1]
        elif isinstance(features[0], str):
            X = data[features]
        else:
            X = data.iloc[:, features]

        if target is None:
            y = data.iloc[:, -1]
        elif isinstance(target, str):
            y = data[[target]]
        else:
            y = data.iloc[:, target]

        X = X.to_numpy()
        y = y.to_numpy()

    else:  # Numpy
        if features is None:
            X = data[:, :-1]
        else:
            X = data[:, np.asarray(features, dtype=np.int64)]

        if target is None:
            y = data[:, -1]
        else:
            y = data[:, np.asarray(target, dtype=np.int64)]

    return convert_to_np_obj(X), convert_to_np_obj(y)


def get_cat_mask(data: npt.NDArray) -> List[int]:
    num_cat_threshold: int = 10
    high_cardinality_threshold: int = 100
    mask: List[int] = []  # True -> cat ; False -> num
    tmp_df: pd.DataFrame = pd.DataFrame(data).infer_objects()
    for col in range(tmp_df.shape[-1]):
        if isinstance(
            tmp_df.iloc[-1, col],
            (int, float, np.int32, np.int64, np.float32, np.float64),
        ):
            if len(tmp_df.iloc[:, col].unique().tolist()) > num_cat_threshold:
                mask.append(0)
            else:
                mask.append(1)
        else:
            if len(tmp_df.iloc[:, col].unique().tolist()) > high_cardinality_threshold:
                mask.append(2)
            else: 
                mask.append(1)
    return mask # 0 - numerical, 1 - cat low cardinality, 2 - cat high cardinality


def calculate_model_score(y: npt.NDArray, y_hat: npt.NDArray, task: str) -> float:
    if task == "tabular_classification":
        return balanced_accuracy_score(y.astype(np.str_), y_hat)
    else:
        score = r2_score(y, y_hat)
        if score < 0:
            score = 0
        score = (score + 1) / 2
        return score


def tab_cv_score(
    pipeline: Pipeline, X: npt.NDArray, y: npt.NDArray, task: str, n_folds: int = 5
) -> List[float]:
    if task == "tabular_classification":
        kf = RepeatedStratifiedKFold(n_splits=n_folds, n_repeats=1)
        y = y.astype(np.str_)
    else:
        kf = RepeatedKFold(n_splits=n_folds, n_repeats=1)
    scores = []
    for train_index, test_index in kf.split(X, y):
        copied_pipeline = deepcopy(pipeline)
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]
        copied_pipeline.fit(X_train, y_train)
        pred = copied_pipeline.predict(X_test)
        scores.append(calculate_model_score(y_test, pred, task))
    return scores



from __future__ import annotations
from abc import ABC, abstractmethod
from numpy import typing as npt
from .task_pipeline import Pipeline
from typing import Dict, Optional, Any, Callable, Type
from falcon.serialization import SerializedModelRepr

class TaskManager(ABC):
    """
    Base class for all Task Managers.
    """
    def __init__(
        self,
        task: str,
        data: Any,
        pipeline: Optional[Type[Pipeline]] = None,
        pipeline_options: Optional[Dict] = None,
        extra_pipeline_options: Optional[Dict] = None,
        features: Any = None,
        target: Any = None,
    ):
        """

        Parameters
        ----------
        task : str
            current task
        data : Any
            data to be used for training
        pipeline : Optional[Type[Pipeline]], optional
            pipeline class to be used, by default None
        pipeline_options : Optional[Dict], optional
            arguments to be passed to pipeline instead of default ones, by default None
        extra_pipeline_options : Optional[Dict], optional
            arguments to be passed to pipeline in addition to default ones, by default None
        features : Any, optional
            featrues to be used for training, by default None
        target : Any, optional
            targets to be used for training, by default None
        """
        self.task: str = task
        self.features = features
        self.target = target
        self._data = self._prepare_data(data)
        self._extra_pipeline_options = extra_pipeline_options
        self._create_pipeline(pipeline=pipeline, options=pipeline_options)

    @abstractmethod
    def train(self, **kwargs: Any) -> TaskManager:
        """
        Trains the underlying pipeline.

        Returns
        -------
        TaskManager
            self
        """
        pass

    @abstractmethod
    def _prepare_data(self, data: Any) -> Any:
        """
        Initial data preparation (e.g. reading from file).
        Warning: initial data preparation (e.g. reading, cleaning) and data preprocessing (e.g. scaling, encoding) are two distinct steps. The later one is performed inside the pipeline.

        Parameters
        ----------
        data : Any
            training data

        Returns
        -------
        Any
            prepared data
        """
        pass

    @property
    @abstractmethod
    def default_pipeline(self) -> Type[Pipeline]:
        """
        Default pipeline class. Can be chosen dynamically.
        """
        pass

    @property
    def default_pipeline_options(self) -> Dict:
        """
        Default pipeline options. Can be chosen dynamically.
        """
        pass

    def predict(self, X: Any) -> Any:
        """
        Calls predict methods of the pipeline.

        Parameters
        ----------
        X : Any
            features

        Returns
        -------
        Any
            predictions
        """
        return self._pipeline.predict(X)

    def _create_pipeline(
        self, pipeline: Optional[Type[Pipeline]], options: Optional[Dict]
    ) -> None:
        """
        Initializes the pipeline.

        Parameters
        ----------
        pipeline : Optional[Type[Pipeline]]
            pipeline class
        options : Optional[Dict]
            pipeline options
        """

        # if pipeline is not None and options is None:
        #     self._pipeline = pipeline(task=self.task)
        if pipeline is None:
            pipeline = self.default_pipeline
        if options is None:
            options = self.default_pipeline_options
            if self._extra_pipeline_options is not None:
                for k, v in self._extra_pipeline_options.items():
                    options[k] = v
        self._pipeline = pipeline(task=self.task, **options)

    def save_model(self, format: str = "auto", filename: Optional[str] = None) -> bytes:
        """
        Serializes and saves the model.

        Parameters
        ----------
        format : str, optional
            "auto", "onnx" or "falcon"; "falcon" format should only be used in rare cases when converting to onnx is not possible, by default "auto"
        filename : Optional[str], optional
            filename for the model file, by default None. If filename is not specified, the model is not saved on disk and only returned as bytes object

        Returns
        -------
        bytes
            serialized model as bytes
        """
        if format not in {"falcon", "onnx", "auto"}:
            raise ValueError(
                f"expected one of [onnx, falcon] as output format, got {format}"
            )
        serialized_model, format = self._pipeline.save(format=format)
        if filename is not None:
            if not filename.endswith(f".{format}"):
                filename += f".{format}"
            with open(filename, "wb+") as f:
                f.write(serialized_model)
        return serialized_model

    @abstractmethod
    def evaluate(self, test_data: Any) -> Any:
        """
        Evaluates the performance of a trained pipeline.

        Parameters
        ----------
        test_data : Any
            data to be used for evaluation
            

        Returns
        -------
        Any
            evaluation metric or None
        """
        pass

    @abstractmethod
    def performance_summary(self, test_data: Any) -> Any: 
        """
        Prints the performance summary of the trained pipeline.
    

        Parameters
        ----------
        test_data : Any
            test set, optional

        Returns
        -------
        Any
            relevant metrics or None
        """
        pass


from typing import List, Optional, Dict
from falcon.utils import print_
import bson
from bson import BSON
from typing import List, Tuple, Optional
from numpy import typing as npt
from onnx import ModelProto, load_from_string
import onnxruntime as ort
from onnx.compose import add_prefix, merge_models
from onnx.helper import make_model
from falcon.config import ONNX_OPSET_VERSION, ML_ONNX_OPSET_VERSION
import numpy as np
from typing import Any, Dict, Union
import onnx
from onnx import TensorProto, helper as h, OperatorSetIdProto

class SerializedModelRepr:
    def __init__(
        self,
        model: bytes,
        n_inputs: int,
        n_outputs: int,
        initial_types: List[str],
        initial_shapes: List[List[Optional[int]]],
        type_: str = "onnx",
    ):
        self._model = model
        self._n_inputs = n_inputs
        self._n_outputs = n_outputs
        self._initial_types = initial_types
        self._initial_shapes = initial_shapes
        self._type = type_

    def get_model(self) -> bytes:
        return self._model

    def get_n_inputs(self) -> int:
        return self._n_inputs

    def get_n_outputs(self) -> int:
        return self._n_outputs

    def get_initial_types(self) -> List[str]:
        return self._initial_types

    def get_initial_shapes(self) -> List[List[Optional[int]]]:
        return self._initial_shapes

    def get_type(self) -> str:
        return self._type

    def to_dict(self) -> Dict:
        return {
            "n_inputs": self._n_inputs,
            "n_outputs": self._n_outputs,
            "initial_types": self._initial_types,
            "initial_shapes": self._initial_shapes,
            "model": self._model,
            "type": self._type,
        }

def serialize_to_onnx(models_: List[SerializedModelRepr]) -> onnx.ModelProto:
    print_("Serializing to onnx...")
    if len(models_) == 0:
        raise ValueError("List of models cannot be empty")

    # Updating the models by resetting the opset and adding prefix to node names
    updated_models: List[ModelProto] = []
    models = [load_from_string(m.get_model()) for m in models_]
    for i, model in enumerate(models):
        op1 = h.make_operatorsetid("", ONNX_OPSET_VERSION)
        op2 = h.make_operatorsetid("ai.onnx.ml", ML_ONNX_OPSET_VERSION)
        updated_model = make_model(model.graph, opset_imports=[op1, op2])
        updated_model = add_prefix(updated_model, prefix=f"falcon_pl_{i}/")
        updated_models.append(updated_model)

    # Merging the models sequentially 
    prev: ModelProto = updated_models[0]
    for i in range(1, len(updated_models)):
        current: ModelProto = updated_models[i]
        prev_outputs: List = prev.graph.output
        current_inputs: List = current.graph.input
        if len(prev_outputs) > len(current_inputs):
            prev_outputs = prev_outputs[: len(current_inputs)]
        if len(prev_outputs) < len(current_inputs):
            raise RuntimeError(
                "When merging, previous model should have at least as many outputs as inputs in the next model."
            )
        io_map: List[Tuple[str, str]] = []
        outputs: List[str] = [o.name for o in current.graph.output]
        for p, c in zip(prev_outputs, current_inputs):
            mapping: Tuple[str, str] = (p.name, c.name)
            io_map.append(mapping)

        print_(f"\t -> Merging step {i} ::: io_map {io_map} ::: outputs: {outputs}")
        combined_model: ModelProto = merge_models(
            prev, current, io_map=io_map  # , outputs=outputs
        )

        prev = combined_model
    print_("Serialization completed.")
    return prev


def serialize_to_falcon(models: List[SerializedModelRepr]) -> bytes:
    print_("Serializing to falcon...")
    if len(models) == 0:
        raise ValueError("List of models cannot be empty")

    nodes = []
    for model_ in models:
        nodes.append(model_.to_dict())

    model_to_save = {"version": 1, "n_nodes": len(nodes), "nodes": nodes}
    encoded = bytes(bson.BSON.encode(model_to_save))
    print_("Serialization completed.")
    return encoded


from numpy import typing as npt
from typing import Dict
import numpy as np
from sklearn import metrics
from falcon.tabular.utils import calculate_model_score

def scale_acc(acc: float, n_classes: int) -> float: 
    if acc < 0. or acc > 1.:
        raise ValueError('Accuracy score should be in range [0,1]')
    elif acc == 0. or acc == 1. or n_classes < 3: 
        return acc

    random_performance = 1 / n_classes

    a_l = 0.5 / random_performance

    a_u = 0.5 / (1-random_performance) 
    b_u = 0.5 - (0.5/(1-random_performance))*random_performance
 
    if acc <= random_performance: 
        return acc * a_l
    else:
        return acc * a_u + b_u

def print_classification_report(y: npt.NDArray, y_hat: npt.NDArray, silent: bool = False) -> Dict:
    y = y.astype(np.str_)
    classification_report = metrics.classification_report(y, y_hat, output_dict=True)
    n_classes, n_samples = len(np.unique(y)), len(y)
    metrics_ = {
        'N_SAMPLES': n_samples, 
        'N_CLASSES': n_classes,
        'ACC': metrics.accuracy_score(y, y_hat),
        'BACC': metrics.balanced_accuracy_score(y, y_hat),
        'PRECISION': list(classification_report[list(classification_report.keys())[-2]].values())[0], 
        'RECALL': list(classification_report[list(classification_report.keys())[-2]].values())[1], 
        'F1': list(classification_report[list(classification_report.keys())[-2]].values())[2], 
        'B_PRECISION': list(classification_report[list(classification_report.keys())[-1]].values())[0],
        'B_RECALL': list(classification_report[list(classification_report.keys())[-1]].values())[1], 
        'B_F1': list(classification_report[list(classification_report.keys())[-1]].values())[2]
    }
     
    metrics_['SCORE'] = metrics_['BACC']
    metrics_['SC_SCORE'] = scale_acc(metrics_['SCORE'], n_classes)
    
    if not silent: 
        confusion_matrix = metrics.confusion_matrix(y, y_hat)
        print()
        print("PERFORMANCE REPORT: CLASSIFICATION")
        print()
        print()
        labels = [
            "Precision (% of correct predictions for this class):",
            "Recall (% of samples of this class that are correctly predicted):",
            "F1 Score (out of samples of this class, % of correct predictions):",
            "Support (number of class samples):",
        ]
        print("CLASS-SPECIFIC METRICS")
        print()
        for k in list(classification_report.keys())[:-3]:
            print(f"Label: {k}")
            print()
            for i, kk in enumerate(classification_report[k].keys()):
                print(labels[i])
                print(classification_report[k][kk])
            print()
        print()
        print("AVERAGE METRICS")
        print()
        print("Confusion Matrix")
        print("Note: high values in diagonal, low values elsewhere indicate good performance")
        print(confusion_matrix)
        print()
        print("Accuracy")
        print(
            "Note: misleading for imbalanced datasets! Low accuracy on classes with a low number of samples is not reflected!"
        )
        print(metrics_['ACC'])
        print()
        print("Balanced Accuracy")
        print("Each class weighs the same even if it has a low number of samples")
        print(metrics_['BACC'])
        print()
        # TODO 
        # print("Weighted ROC AUC score")
        # print(
        #     "Area Under the Receiver Operating Characteristic Curve, balanced by number of samples per class. The closer to 1 the better the performance."
        # )
        # print(metrics.roc_auc_score(np.eye(np.max(y) + 1)[y], np.eye(np.max(y_hat) + 1)[y_hat], multi_class="ovo"))
        # print()
        print("Precision")
        print("Note: misleading for imbalanced datasets!")
        print(metrics_['PRECISION'])
        print()
        print("Recall")
        print("Note: misleading for imbalanced datasets!")
        print(metrics_['RECALL'])
        print()
        print("F1 Score")
        print("Note: misleading for imbalanced datasets!")
        print(metrics_['F1'])
        print()
        print("Balanced Precision")
        print(metrics_['B_PRECISION'])
        print()
        print("Balanced Recall")
        print(metrics_['B_RECALL'])
        print()
        print("Balanced F1 Score")
        print(metrics_['B_F1'])
    return metrics_


def print_regression_report(y: npt.NDArray, y_hat: npt.NDArray, silent: bool = False) -> Dict:
    diff = y-y_hat
    metrics_ = {
        'N_SAMPLES': len(y),
        'R2': metrics.r2_score(y, y_hat), 
        'RMSE': np.sqrt(np.mean((diff) ** 2)), 
        'MSE': np.mean((diff) ** 2), 
        'MAE': np.mean(np.abs(diff)),
        'RMSLE': np.log(np.sqrt(np.mean((diff) ** 2)) + 1e-7)
    }
    metrics_['SCORE'] = metrics_['R2'] if metrics_['R2'] > 0.0 else 0.0 
    metrics_['SC_SCORE'] = (metrics_['SCORE'] + 1) / 2
    if not silent:
        print("PERFORMANCE REPORT: REGRESSION")
        print()
        print("R2")
        print(
            "The closer to 1 the better the performance. R2 of 0 is the score of a regressor that always predicts the average."
        )
        print(metrics_['R2'])
        print()
        print("RMSE (Root Mean Squared Error)")
        print(metrics_['RMSE'])
        print()
        print("MSE (Mean Sqaured Error)")
        print(metrics_['MSE'])
        print()
        print("MAE (Mean Absolute Error)")
        print(metrics_['MAE'])
        print()
        print("RMSLE (Root Mean Squared Log Error)")
        print("Useful in case of skewed distribution of target values.")
        print(metrics_['RMSLE'])
    return metrics_



from sklearn.ensemble import (
    StackingRegressor as SklearnStackingRegressor,
    StackingClassifier as SklearnStackingClassifier,
)
from falcon.abstract.onnx_convertible import ONNXConvertible
from falcon.addons.sklearn import (
    BalancedStackingClassifier as SklearnBalancedStackingClassifier,
)
from falcon.abstract import Model
from falcon.types import Float32Array
from sklearn.base import BaseEstimator
from typing import Any, List, Optional, Union, Dict, Tuple, Callable, Type
from numpy import typing as npt
import numpy as np
from skl2onnx import convert_sklearn
from skl2onnx.common.data_types import TensorType, FloatTensorType
from falcon.config import ONNX_OPSET_VERSION, ML_ONNX_OPSET_VERSION
from falcon.serialization import SerializedModelRepr


class _StackingBase(Model, ONNXConvertible):
    def __init__(
        self,
        cls: Callable,
        estimators: List[Tuple[str, BaseEstimator]],
        final_estimator: BaseEstimator,
        cv: Any = None,
        n_jobs: int = -1,
        passthrough: bool = False,
        verbose: int = 0,
        **kwargs: Any,
    ) -> None:
        self.estimator: BaseEstimator = cls(
            estimators=estimators,
            final_estimator=final_estimator,
            cv=cv,
            n_jobs=n_jobs,
            passthrough=passthrough,
            **kwargs,
        )

        self._ret_type: Type = (
            np.float32
            if isinstance(self.estimator, SklearnStackingRegressor)
            else np.int64
        )

    def fit(self, X: Float32Array, y: Float32Array) -> None:
        """
        Fits the model

        Parameters
        ----------
        X : Float32Array
            Features
        y : Float32Array
            targets
        """
        self._shape = [None, *X.shape[1:]]
        self.estimator.fit(X, y)

    def predict(self, X: Float32Array) -> Float32Array:
        """
        Predicts the target for the given input

        Parameters
        ----------
        X : Float32Array
            featrues

        Returns
        -------
        Float32Array
            predictions
        """
        prediction: npt.NDArray = self.estimator.predict(X)
        prediction = prediction.astype(dtype=self._ret_type)
        return prediction

    def to_onnx(self) -> SerializedModelRepr:
        """
        Serializes the model to onnx. 

        Returns
        -------
        SerializedModelRepr
        """
        initial_type = [("model_input", FloatTensorType(self._shape))]
        options = self._get_onnx_options()
        onnx_model = convert_sklearn(
            self.estimator,
            initial_types=initial_type,
            target_opset={'': ONNX_OPSET_VERSION, 'ai.onnx.ml': ML_ONNX_OPSET_VERSION},
            # final_types=self._get_onnx_final_types(),
            options=options,
        )
        n_inputs = len(onnx_model.graph.input)
        n_outputs = len(onnx_model.graph.output)

        return SerializedModelRepr(
            onnx_model.SerializeToString(),
            n_inputs,
            n_outputs,
            ["FLOAT32"],
            [self._shape],
        )

    # def _get_onnx_final_types(self) -> List[Tuple[str, TensorType]]:
    #     return [("stacking_output", FloatTensorType([None, 1]))]

    def _get_onnx_options(self) -> Dict:
        return {}


class StackingClassifier(_StackingBase):
    def __init__(
        self,
        estimators: List[Tuple[str, BaseEstimator]],
        final_estimator: BaseEstimator,
        balanced: bool = True,
        cv: Any = 5,
        n_jobs: int = -1,
        passthrough: bool = False,
        verbose: int = 0,
        stack_method: Any = "auto",
        **kwargs: Any,
    ) -> None:
        """
        Small wrapper around `sklearn.ensemble.StackingClassifier`. 
        For more detailed description please refer to sklarn documentation. 


        Parameters
        ----------
        estimators : List[Tuple[str, BaseEstimator]]
            base estimators
        final_estimator : BaseEstimator
            meta estimator, by default LogisticRegression
        balanced : bool, optional
            if True, the classes are balanced by performing random oversampling, by default True
        cv : Any, optional
            number of CV folds, or custom CV object, by default 5
        n_jobs : int, optional
            number of parallel jobs, by default -1
        passthrough : bool, optional
            when True the meta estimator is trained on original data in addition to the predictions of base estimators, by default False
        verbose : int, optional
            verbosity level of underlying sklearn estimator, by default 0
        stack_method : Any, optional
            methods called for each base estimator, by default "auto"
        """
        cls: Callable
        if balanced:
            cls = SklearnBalancedStackingClassifier
        else:
            cls = SklearnStackingClassifier
        super().__init__(
            cls=cls,
            estimators=estimators,
            final_estimator=final_estimator,
            cv=cv,
            n_jobs=n_jobs,
            passthrough=passthrough,
            verbose=verbose,
            stack_method=stack_method,
            **kwargs,
        )

    # def _get_onnx_final_types(
    #     self,
    # ) -> List[Tuple[str, TensorType]]:
    #     return [
    #         ("stacking_labels", FloatTensorType([None])),
    #         ("stacking_probs", FloatTensorType([None, None])),
    #     ]

    def _get_onnx_options(self) -> Dict:
        return {id(self.estimator): {"zipmap": False}}


class StackingRegressor(_StackingBase):
    def __init__(
        self,
        estimators: List[Tuple[str, BaseEstimator]],
        final_estimator: BaseEstimator,
        cv: Any = 5,
        n_jobs: int = -1,
        passthrough: bool = False,
        verbose: int = 0,
        **kwargs: Any,
    ) -> None:
        """
        Small wrapper around `sklearn.ensemble.StackingRegressor`. 
        For more detailed description please refer to sklarn documentation. 

        Parameters
        ----------
        estimators : List[Tuple[str, BaseEstimator]]
            base estimators
        final_estimator : BaseEstimator
            meta estimator
        cv : Any, optional
            number of CV folds, or custom CV object, by default 5
        n_jobs : int, optional
            number of parallel jobs, by default -1
        passthrough : bool, optional
            when True the meta estimator is trained on original data in addition to the predictions of base estimators, by default False
        verbose : int, optional
            verbosity level of underlying sklearn estimator, by default 0
        """
        super().__init__(
            cls=SklearnStackingRegressor,
            estimators=estimators,
            final_estimator=final_estimator,
            cv=cv,
            n_jobs=n_jobs,
            passthrough=passthrough,
            verbose=verbose,
            **kwargs,
        )



from typing import Union, List, Tuple, Optional
import numpy as np
from numpy import typing as npt


ColumnsList = Union[List[str], List[int]]
Float32Array = npt.NDArray[np.float32]
Int64Array = npt.NDArray[np.int64]



from numpy import typing as npt
import numpy.typing as npt
from typing import List, Any, Optional, Dict, Type
from falcon.abstract import Pipeline, PipelineElement
from falcon.abstract.learner import Learner
from falcon.abstract.onnx_convertible import ONNXConvertible
from falcon.tabular.processors.label_decoder import LabelDecoder
from falcon.tabular.processors.scaler_and_encoder import ScalerAndEncoder
from falcon.tabular.learners.super_learner import SuperLearner
from falcon.utils import print_

class SimpleTabularPipeline(Pipeline):
    """
    Default tabular pipeline.
    """
    def __init__(self, task: str, mask: List[int], learner: Type[Learner] = SuperLearner, learner_kwargs: Optional[Dict] = None):
        """
        Default tabular pipeline. On a high level it simply chains a preprocessor and model learner (by default `SuperLearner`). 
        For classification tasks, the labels are also encoded as integers (while predictions are decoded back to strings).
        Internally, all numerical features are scaled to 0 mean and 1 std. All categorical features are one-hot encoded (this approach might not be suitable for features with very high cardinality). 

        Parameters
        ----------
        task : str
            `tabular_classification` or `tabular_regression`
        mask : List[int]
            list of ints where 1/2 indicates a low/high cardinality categorical feature and 0 indicates a numerical feature
        learner : Learner, optional
            learner class to be used, by default `SuperLearner`
        learner_kwargs : Optional[Dict], optional
            arguments to be passed to the learner, by default None
        """
        super().__init__(task=task)
        
        encoder: PipelineElement = ScalerAndEncoder(mask)
        self.add_element(encoder)
        
        if not learner_kwargs:
            learner_kwargs = {}
        learner_: PipelineElement = learner(task=task, **learner_kwargs)
        self.add_element(learner_)

        if self.task == "tabular_classification":
            self.labels_transformer: LabelDecoder = LabelDecoder()
            self.add_element(self.labels_transformer)

    def fit(self, X: npt.NDArray, y: npt.NDArray) -> None:
        """
        Fits the pipeline by consecutively calling `.fit_pipe()` method of each element in pipeline.
        For tabular classification, `LabelDecoder` is applied to targets before actual training occurs.

        Parameters
        ----------
        X : npt.NDArray
            train featrues
        y : npt.NDArray
            train targets
        """
        print_('Fitting the pipeline...')
        if self.task == "tabular_classification":
            self.labels_transformer.fit(y)
            y = self.labels_transformer.transform(y, inverse=False)
        for p in self._pipeline:
            p.fit_pipe(X, y)
            X = p.forward(X)

    def predict(self, X: npt.NDArray) -> npt.NDArray:
        """
        Predicts the label of passed data points.

        Parameters
        ----------
        X : npt.NDArray
            features

        Returns
        -------
        npt.NDArray
            predicted label
        """
        for p in self._pipeline:
            X = p.forward(X)
        return X

    

    


from sklearn.model_selection import StratifiedKFold
from imblearn.over_sampling import RandomOverSampler


class BalancedStratifiedKFold(StratifiedKFold):
    def split(self, X, y, groups=None): # type: ignore
        for train, test in super().split(X, y, groups):
            y_train_split = y[train]
            y_test_split = y[test]
            train, _ = RandomOverSampler().fit_resample(
                train.reshape(-1, 1), y_train_split
            )

            
            yield train.squeeze(), test.squeeze()



# type: ignore
from imblearn.over_sampling import RandomOverSampler
from abc import ABCMeta, abstractmethod
from copy import deepcopy

import numpy as np
from joblib import Parallel
import scipy.sparse as sparse

from sklearn.base import clone
from sklearn.base import is_classifier
from sklearn.ensemble._base import _fit_single_estimator
from sklearn.model_selection import cross_val_predict
from sklearn.model_selection import check_cv
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import Bunch
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import check_scalar
from sklearn.utils.fixes import delayed
from sklearn.ensemble import StackingClassifier
from types import MethodType
from falcon.addons.sklearn.model_selection.balanced_strat_kfold import (
    BalancedStratifiedKFold,
)

# Slightly modified version of StackingClassifier from sklearn that upsamples the minority class during training
# The .fit() method was adopted from https://github.com/scikit-learn/scikit-learn/blob/36958fb24/sklearn/ensemble/_stacking.py


def _fit(self, X, y, sample_weight=None):
    check_classification_targets(y)
    self._le = LabelEncoder().fit(y)
    self.classes_ = self._le.classes_
    check_scalar(
        self.passthrough,
        name="passthrough",
        target_type=(np.bool_, bool),
        include_boundaries="neither",
    )
    # all_estimators contains all estimators, the one to be fitted and the
    # 'drop' string.
    names, all_estimators = self._validate_estimators()
    self._validate_final_estimator()

    stack_method = [self.stack_method] * len(all_estimators)

    if self.cv == "prefit":
        self.estimators_ = []
        for estimator in all_estimators:
            if estimator != "drop":
                check_is_fitted(estimator)
                self.estimators_.append(estimator)
    else:
        # Fit the base estimators on the whole training data. Those
        # base estimators will be used in transform, predict, and
        # predict_proba. They are exposed publicly.
        X_resampled, y_resampled = RandomOverSampler().fit_resample(X, y)
        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_single_estimator)(
                clone(est), X_resampled, y_resampled, sample_weight
            )
            for est in all_estimators
            if est != "drop"
        )

    self.named_estimators_ = Bunch()
    est_fitted_idx = 0
    for name_est, org_est in zip(names, all_estimators):
        if org_est != "drop":
            current_estimator = self.estimators_[est_fitted_idx]
            self.named_estimators_[name_est] = current_estimator
            est_fitted_idx += 1
            if hasattr(current_estimator, "feature_names_in_"):
                self.feature_names_in_ = current_estimator.feature_names_in_
        else:
            self.named_estimators_[name_est] = "drop"

    self.stack_method_ = [
        self._method_name(name, est, meth)
        for name, est, meth in zip(names, all_estimators, stack_method)
    ]

    if self.cv == "prefit":
        # Generate predictions from prefit models
        predictions = [
            getattr(estimator, predict_method)(X)
            for estimator, predict_method in zip(all_estimators, self.stack_method_)
            if estimator != "drop"
        ]
    else:
        # To train the meta-classifier using the most data as possible, we use
        # a cross-validation to obtain the output of the stacked estimators.
        # To ensure that the data provided to each estimator are the same,
        # we need to set the random state of the cv if there is one and we
        # need to take a copy.

        if isinstance(self.cv, int):
            self.cv = BalancedStratifiedKFold(self.cv)

        cv = check_cv(self.cv, y=y, classifier=is_classifier(self))
        if hasattr(cv, "random_state") and cv.random_state is None:
            cv.random_state = np.random.RandomState()

        fit_params = (
            {"sample_weight": sample_weight} if sample_weight is not None else None
        )

        predictions = Parallel(n_jobs=self.n_jobs)(
            delayed(cross_val_predict)(
                clone(est),
                X,
                y,
                cv=deepcopy(cv),
                method=meth,
                n_jobs=self.n_jobs,
                fit_params=fit_params,
                verbose=self.verbose,
            )
            for est, meth in zip(all_estimators, self.stack_method_)
            if est != "drop"
        )

    # Only not None or not 'drop' estimators will be used in transform.
    # Remove the None from the method as wl.
    self.stack_method_ = [
        meth for (meth, est) in zip(self.stack_method_, all_estimators) if est != "drop"
    ]

    X_meta = self._concatenate_predictions(X, predictions)
    X_meta_resampled, y_meta_resampled = RandomOverSampler().fit_resample(X_meta, y)
    _fit_single_estimator(
        self.final_estimator_,
        X_meta_resampled,
        y_meta_resampled,
        sample_weight=sample_weight,
    )

    return self


# the object is being patched with a new method instead of subclassing
# so the estimator can be converted to ONNX using the default converter
def BalancedStackingClassifier(estimators, final_estimator, **kwargs):
    clf = StackingClassifier(estimators, final_estimator, **kwargs)
    clf.fit = MethodType(_fit, clf)
    return clf



from abc import ABC, abstractmethod
from numpy import typing as npt
from typing import Any


class Model(ABC):
    """
    Base class for all models.
    """

    @abstractmethod
    def fit(self, X: npt.NDArray, y: npt.NDArray) -> Any:
        """

        Parameters
        ----------
        X : npt.NDArray
            features
        y : npt.NDArray
            targets

        Returns
        -------
        Any
            usually `None`
        """
        pass

    @abstractmethod
    def predict(self, X: npt.NDArray) -> npt.NDArray:
        """
        Parameters
        ----------
        X : npt.NDArray
            features

        Returns
        -------
        npt.NDArray
            predictions
        """
        pass


class TransformerMixin:
    def transform(self, X: npt.NDArray) -> npt.NDArray:
        """
        Equivalent of `self.predict(X)`

        Parameters
        ----------
        X : npt.NDArray
            features

        Returns
        -------
        npt.NDArray
            transformed features
        """
        return self.predict(X)  # type: ignore



from __future__ import annotations
from falcon.abstract import TaskManager, Pipeline
from falcon.tabular.pipelines.simple_tabular_pipeline import SimpleTabularPipeline
from falcon.tabular.utils import convert_to_np_obj
from .reporting import print_classification_report, print_regression_report
from falcon.tabular.utils import *
from falcon import types as ft
from typing import Union, Optional, List, Tuple, Type, Dict, Any
from numpy import typing as npt
import pandas as pd
from falcon.utils import print_, set_verbosity_level
from sklearn.model_selection import train_test_split
import os
import pandas as pd

class TabularTaskManager(TaskManager):
    """
    Default task manager for tabular data.
    """

    def __init__(
        self,
        task: str,
        data: Union[str, npt.NDArray, pd.DataFrame, Tuple], 
        pipeline: Optional[Type[Pipeline]] = None,
        pipeline_options: Optional[Dict] = None,
        extra_pipeline_options: Optional[Dict] = None,
        features: Optional[ft.ColumnsList] = None,
        target: Optional[Union[str, int]] = None,
        **options: Any,
    ) -> None:
        """

        Parameters
        ----------
        task : str
            `tabular_classification` or `tabular_regression`
        data : Union[str, npt.NDArray, pd.DataFrame, Tuple]
            path to data file or pandas dataframe or numpy array or tuple (X,y)
        pipeline: Optional[Type[Pipeline]] 
            class to be used as pipeline, by default None.
            If None, `SimpleTabularPipeline` will be used
        pipeline_options : Optional[Dict], optional
            arguments to be passed to the pipeline, by default None.
            These options will overwrite the ones from `default_pipeline_options` attribute.
        extra_pipeline_options : Optional[Dict], optional
            arguments to be passed to the pipeline, by default None.
            These options will be passed in addition to the ones from `default_pipeline_options` attribute.
            This argument is ignored if `pipeline_options` is not None
        features : Optional[ft.ColumnsList], optional
            names or indices of columns to be used as features, by default None.
            If None, all columns except the last one will be used.
            If `target` argument is not None, features should be passed explicitly as well
        target : Optional[Union[str, int]], optional
            name or index of column to be used as target, by default None.
            If None, the last column will be used as target.
            If `features` argument is not None, target should be specified explicitly as well
        """
        print_(f"\nInitializing a new TabularTaskManager for task `{task}`")

        super().__init__(
            task=task,
            data=data,
            pipeline=pipeline,
            pipeline_options=pipeline_options,
            extra_pipeline_options=extra_pipeline_options,
            features=features,
            target=target,
        )

        self._eval_set: Optional[Tuple] = None

    def _prepare_data(
        self, data: Union[str, npt.NDArray, pd.DataFrame, Tuple], training: bool = True
    ) -> Tuple[npt.NDArray, npt.NDArray, List[int]]:
        """
        Initial data preparation: 
        1) optional: read data from the specified location;
        2) split into features and targets. By default it is assumed that the last column is the target;
        3) clean data;
        4) determine numerical and categorical features (create categorical mask).

        Parameters
        ----------
        data : Union[str, npt.NDArray, pd.DataFrame, Tuple]
            path to data file or pandas dataframe or numpy array or Tuple(X,y)

        Returns
        -------
        Tuple[npt.NDArray, npt.NDArray, List[int]]
            tuple of features, target and categorical mask for features
        """
        if isinstance(data, str):
            data = read_data(data)
        if isinstance(data, tuple):
            if self.features is not None or self.target is not None: 
                print('When data is passed as tuple of (X, y) all columns are used regardless the values of `features` or `target` arguments.')
            if len(data) != 2: 
                raise ValueError('When passing data as tuple, it should contain exactly 2 elements: `X` and `y`.')
            X, y = data
            if len(y.shape) > 3 or len(X.shape) > 3: 
                raise ValueError('Invalid data shape.')
            if len(y.shape) > 1 and y.shape[-1] != 1: 
                raise ValueError('The target should contain only one column.')
            X, y = convert_to_np_obj(X), convert_to_np_obj(y)
        else:
            X, y = split_features(data, features=self.features, target=self.target)
        X, y = clean_data_split(X, y)
        mask: List[int]
        if training:
            mask = get_cat_mask(X)
        else: 
            mask = []
        if len(y.shape) == 2:
            y = y.ravel()
        return X, y, mask

    @property
    def default_pipeline(self) -> Type[Pipeline]:
        """
        Default pipeline class.
        """

        return SimpleTabularPipeline

    @property
    def default_pipeline_options(self) -> Dict:
        """
        Default options for pipeline.
        """
        options = {"mask": self._data[2]}
        return options
    
    def _pre_eval_pipeline(self) -> float: 
        print_("Pre-evaluation of model performance")
        # TODO make `no_print` context manager
        old_verbosity_level = int(os.getenv("FALCON_VERBOSITYLEVEL", "1"))
        set_verbosity_level(0)
        if (
            self._data[0].shape[0] * self._data[0].shape[1] < 50000
            or self._data[0].shape[0] < 500
        ):
            score = tab_cv_score(
                self._pipeline, self._data[0], self._data[1], self.task
            )
            avg_score = float(np.mean(score))
        else:
            X_train, X_test, y_train, y_test = train_test_split(
                self._data[0], self._data[1], test_size=0.25, stratify = self._data[1] if self.task == 'tabular_classification' else None
            )
            copied_pipeline = deepcopy(self._pipeline)
            copied_pipeline.fit(X_train, y_train)
            pred = copied_pipeline.predict(X_test)
            avg_score = calculate_model_score(y_test, pred, self.task)
        set_verbosity_level(old_verbosity_level)
        return avg_score


    def train(self, make_eval_subset: bool = True, pre_eval: bool = False, **kwargs: Any) -> TabularTaskManager:
        """
        Invokes the training procedure of an underlying pipeline. Print an expected model performance if available.

        Parameters
        ----------
        pre_eval : bool
            if True, first estimate model perfromance via 10 folds CV for small datasets or 25% test split for large datasets, by default False.
            Setting pre_eval = True is not reccomended as it pre-evaluates the pipeline as a whole which has lots of random elements therefore the results might be non reproducable
        make_eval_subset: bool
            controls whether a dedicated eval set should be allocated for performance report, by default True. 
            If True, overwrites the value of `pre_eval` to False
        Returns
        -------
        TabularTaskManager
            `self`
        """
        print_("Beginning training")
        if make_eval_subset:
            if self._eval_set is None:
                X_train, X_eval, y_train, y_eval = train_test_split(
                    self._data[0], self._data[1], test_size=0.25, stratify = self._data[1] if self.task == 'tabular_classification' else None
                )
                self._data = (X_train, y_train, self._data[2])
                self._eval_set = (X_eval, y_eval)
            else: 
                print('Evaluation set is already available.')
        elif pre_eval:
            print("Setting `pre_eval = True` is not reccomended as it pre evaluates the whole pipeline, can be lenghty and innacurate")
            preeval_score = self._pre_eval_pipeline()
            print(f"Pre evaluation score: {preeval_score}")
        print_("Beginning the main training phase")
        self._pipeline.fit(self._data[0], self._data[1])
        print_("Finished training")
        return self

    def predict(self, data: Union[str, npt.NDArray, pd.DataFrame]) -> npt.NDArray:
        """
        Performs prediction on new data.

        Parameters
        ----------
        data : Union[str, npt.NDArray, pd.DataFrame]
             path to data file or pandas dataframe or numpy array

        Returns
        -------
        npt.NDArray
            predictions
        """
        if isinstance(data, str):
            data = read_data(data)
        if not isinstance(data, np.ndarray):
            data = np.asarray(data, dtype=np.object_)
        return self._pipeline.predict(data)

    def predict_stored_subset(self, subset: str = 'train') -> npt.NDArray:
        """
        Makes a prediction on a stored subset (`train` or `eval`).

        Parameters
        ----------
        subset : str, optional
            subset to predict on (train or eval), by default 'train'

        Returns
        -------
        npt.NDArray
            predicted values
        """
        if subset == 'train': 
            return self.predict(self._data[0])
        elif subset == 'eval':
            if self._eval_set is None: 
                raise RuntimeError('validation set is not available') 
            return self.predict(self._eval_set[0])
        else:
            raise ValueError('subset should be either `train` or `eval`')

    def performance_summary(self, test_data: Optional[Union[str, npt.NDArray, pd.DataFrame, Tuple]]) -> dict:
        """
        Prints a performance summary of the model. 
        The summary always includes metrics calculated for the train set. 
        If the train/eval split was done during training, the summary includes metrics calculated on eval set. 
        If test set is provided as an argument, the performance includes metrics calculated on test set.

        Parameters
        ----------
        test_data : Optional[Union[str, npt.NDArray, pd.DataFrame, Tuple]]
            data to be used as test set, by default None

        Returns
        -------
        dict
            metrics for each subset
        """
        metrics_ = {}
        report_fn = print_classification_report if self.task == 'tabular_classification' else print_regression_report
        y_hat_train = self.predict_stored_subset('train')
        metrics_['train'] = report_fn(self._data[1], y_hat_train, silent = True)
        if self._eval_set is not None: 
            y_hat_eval = self.predict_stored_subset('eval')
            metrics_['eval'] = report_fn(self._eval_set[1], y_hat_eval, silent = True)
        if test_data is not None: 
            metrics_['test'] = self.evaluate(test_data, silent = True)
        df = pd.DataFrame.from_dict(metrics_, orient = 'index')
        print("\n", df, "\n")
        return metrics_

    def evaluate(self, test_data: Union[str, npt.NDArray, pd.DataFrame, Tuple], silent: bool = False) -> Dict:
        """
        Perfoms and prints the evaluation report on the given dataset.

        Parameters
        ----------
        test_data : Union[str, npt.NDArray, pd.DataFrame, Tuple]
            dataset to be used for evaluation
        silent: bool
            controls whether the metrics are printed on screen, by default False
        """
        print("The evaluation report will be provided here")
        X, y, _ = self._prepare_data(test_data, training = False)
        y_hat = self.predict(X)
        if self.task == "tabular_classification":
            return print_classification_report(y, y_hat, silent = silent)
        else:
            return print_regression_report(y, y_hat, silent = silent)



from abc import ABC, abstractmethod
from numpy import typing as npt
from falcon.abstract.task_pipeline import PipelineElement
from typing import Any, Optional


class Learner(PipelineElement):
    """
    Subclass of `PipelineElement`.
    Learners are task aware pipeline elements that act as wrappers around models and responsible for tuning of the hyperparameters.
    """
    def __init__(self, task: str, **kwargs: Any) -> None: 
        """
        Parameters
        ----------
        task : str
            current ML task
        """
        self.task = task

    


from falcon.tabular.pipelines import SimpleTabularPipeline
from falcon.tabular.learners import SuperLearner
from falcon.tabular.learners.super_learner import _default_estimators

TABULAR_CLASSIFICATION_CONFIGURATIONS = {
    "SuperLearner.mini": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 10, 
                "base_estimators": _default_estimators['tabular_classification']['mini']
            }
        }
    }, 

    "SuperLearner.mid": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 5, 
                "base_estimators": _default_estimators['tabular_classification']['mid']
            }
        }
    },

    "SuperLearner.large": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 3, 
                "base_estimators": _default_estimators['tabular_classification']['large']
            }
        }
    }
}

TABULAR_REGRESSION_CONFIGURATIONS = {
    "SuperLearner.mini": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 10, 
                "base_estimators": _default_estimators['tabular_regression']['mini']
            }
        }
    }, 

    "SuperLearner.mid": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 5, 
                "base_estimators": _default_estimators['tabular_regression']['mid']
            }
        }
    },

    "SuperLearner.large": {
        "pipeline": SimpleTabularPipeline, 
        "extra_pipeline_options": {
            "learner": SuperLearner, 
            "learner_kwargs": {
                "cv": 3, 
                "base_estimators": _default_estimators['tabular_regression']['large']
            }
        }
    }
}


from abc import abstractmethod
from typing import Any, Type, Union, Optional, List, Tuple
from falcon.abstract.model import Model
from falcon.abstract.onnx_convertible import ONNXConvertible
from falcon.serialization import SerializedModelRepr, serialize_to_falcon, serialize_to_onnx


class PipelineElement(Model):
    """
    Base class for all pipeline elements.
    """

    @abstractmethod
    def get_input_type(self) -> Type:
        """
        Returns
        -------
        Type
            Input types
        """
        pass

    @abstractmethod
    def get_output_type(self) -> Type:
        """
        Returns
        -------
        Type
            Output types
        """
        pass

    def forward(self, X: Any) -> Any:
        """
        Equivalent of `predict` method that is used for elements chaining inside pipeline during inference.

        Parameters
        ----------
        X : Any
            featrues

        Returns
        -------
        Any
            predictions
        """
        self.predict(X)

    def fit_pipe(self, X: Any, y: Any) -> Any:
        """
        Equivalent of `fit` method that is used for elements chaining inisde pipeline during training.

        Parameters
        ----------
        X : Any
            features
        y : Any
            targets

        Returns
        -------
        Any
            usually None
        """
        self.fit(X, y)


class Pipeline(Model):
    """
    Base class for all pipelines.
    """

    def __init__(self, task: str, **kwargs: Any) -> None:
        self.task = task
        self._pipeline: List[PipelineElement] = []

    def add_element(self, element: PipelineElement) -> None:
        """
        Adds element to pipeline. The input type of added element should match the output type of the last element in the pipeline.

        Parameters
        ----------
        element : PipelineElement
            element to be added to the end of the pipeline
        """
        if (
            len(self._pipeline) > 1
            and element.get_input_type() != self._pipeline[-1].get_output_type()
        ):
            raise RuntimeError(
                "The element cannot be added to pipeline due to input type missmatch."
            )
        if element is self:
            raise ValueError("Cannot add self to the pipeline")
        self._pipeline.append(element)

    def save(self, format: str = "auto") -> Tuple[bytes, str]:
        """
        Serializes the model to string. For more details please refer to the documentation of `TaskManager.save_model(...)` method.

        Parameters
        ----------
        format : str, optional
           `auto`, `falcon` or `onnx`, by default `auto`

        Returns
        -------
        Tuple[SerializedModelRepr, str]
            Serialized model, format


        """
        if format not in {"falcon", "onnx"}:
            raise ValueError(
                f"expected one of [onnx, falcon] as output format, got {format}"
            )
        serialized_pipeline_elements: List[SerializedModelRepr] = []
        for p in self._pipeline:
            if isinstance(p, ONNXConvertible):
                serialized_pipeline_elements.append(p.to_onnx())
            else:
                if format == "onnx":
                    raise RuntimeError("This model cannot be serialized to onnx")
                elif format == "auto":
                    format = "falcon"
        if format == "falcon":
            serialized_model = serialize_to_falcon(serialized_pipeline_elements)
        else:
            format = "onnx"
            serialized_model = serialize_to_onnx(
                serialized_pipeline_elements
            ).SerializeToString()
        return serialized_model, format



from concurrent.futures import process
from multiprocessing.dummy import Process
from falcon.abstract import Processor, ONNXConvertible, PipelineElement
from typing import Any, Type, Union
from numpy.typing import NDArray
import numpy as np
from sklearn.preprocessing import LabelEncoder
from falcon.types import Float32Array, Int64Array
from skl2onnx import convert_sklearn
from onnx import TensorProto, helper as h, OperatorSetIdProto
from skl2onnx.common.data_types import StringTensorType
from falcon.config import ML_ONNX_OPSET_VERSION
from numpy import typing as npt
from falcon.serialization import SerializedModelRepr


class LabelDecoder(Processor, ONNXConvertible):
    """
    Label encoder/decoder to be used for encoding labels as integers and vice versa.
    """
    def __init__(self) -> None:
        """
        does not take any arguments
        """
        self.le = LabelEncoder()

    def fit_pipe(self, _: Any, __: Any) -> None:  # Do nothing
        """
        Since label decoder should initially be fitted and applied before the main training phase of pipeline, this method does nothing. 

        Parameters
        ----------
        _ : Any
            dummy argument
        __ : Any
            dummy argument
        """
        return

    def fit(self, X: npt.NDArray, _: Any = None) -> None:
        """
        Fits the decoder.

        Parameters
        ----------
        X : npt.NDArray
            labels to be encoded as integers
        _ : Any, optional
            dummy argument, by default None
        """
        self.le.fit(X)

    def predict(self, X: npt.NDArray, inverse: bool = True) -> npt.NDArray:
        """
        Equivalent of `.transform()`.

        Parameters
        ----------
        X : npt.NDArray
            labels
        inverse : bool, optional
            if True, encode strings as integers, else convert integers back to strings, by default True

        Returns
        -------
        npt.NDArray
            encoded/decoded labels
        """
        return self.transform(X, inverse=inverse)

    def transform(self, X: npt.NDArray, inverse: bool = True) -> npt.NDArray:
        """
        Encodes/decodes the labels.

        Parameters
        ----------
        X : npt.NDArray
            labels
        inverse : bool, optional
            if True, encode strings as integers, else convert integers back to strings, by default True

        Returns
        -------
        npt.NDArray
            encoded/decoded labels
        """
        if not inverse:
            return self.le.transform(X)
        else:
            return self.le.inverse_transform(X.astype(np.int64)).astype(np.str_)

    def get_input_type(self) -> Type:
        """
        Returns
        -------
        Type
            Int64Array
        """
        return Int64Array

    def get_output_type(self) -> Type:
        """
        Returns
        -------
        Type
            NDArray[np.str_]
        """
        return NDArray[np.str_]

    def to_onnx(self) -> SerializedModelRepr:
        """
        Serializes the encoder to onnx. 

        Returns
        -------
        SerializedModelRepr
        """
        inputs = [h.make_tensor_value_info("encoded_labels", TensorProto.INT64, [None])]
        outputs = [
            h.make_tensor_value_info("decoded_labels", TensorProto.STRING, [None])
        ]
        node = h.make_node(
            "LabelEncoder",
            ["encoded_labels"],
            ["decoded_labels"],
            values_strings=[str(el) for el in self.le.classes_],
            keys_int64s=[int(i) for i in range(len(self.le.classes_))],
            name=f"labels_decoder",
            domain="ai.onnx.ml",
        )
        graph = h.make_graph([node], f"decoder", inputs, outputs)
        op = h.make_operatorsetid("ai.onnx.ml", ML_ONNX_OPSET_VERSION)
        model = h.make_model(graph, producer_name="falcon", opset_imports = [op])
        return SerializedModelRepr(model.SerializeToString(), 1, 1, ["INT64"], [[None]])

    def forward(
        self, X: npt.NDArray
    ) -> npt.NDArray:  # Inside pipeline used as post-processor to decode labels back to strings
        """
        Equivalent to `.transform(X, inverse=True)`.

        Parameters
        ----------
        X : npt.NDArray
            labels to decode

        Returns
        -------
        npt.NDArray
            labels decoded to strings
        """
        return self.transform(X, inverse=True)



from copy import deepcopy
from typing import Dict
from falcon.tabular.configurations import (
    TABULAR_CLASSIFICATION_CONFIGURATIONS,
    TABULAR_REGRESSION_CONFIGURATIONS,
)


def get_task_configuration(task: str, configuration_name: str) -> Dict:
    if task == "tabular_classification":
        configs = TABULAR_CLASSIFICATION_CONFIGURATIONS
    elif task == "tabular_regression":
        configs = TABULAR_REGRESSION_CONFIGURATIONS
    else:
        raise ValueError("Unknown task")

    if configuration_name not in configs.keys():
        raise ValueError(f"Task {task} has no config with name `{configuration_name}`")
    
    return deepcopy(configs[configuration_name])


import numpy as np
from numpy import typing as npt
from falcon.types import Float32Array
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from falcon.abstract import Processor, ONNXConvertible
from skl2onnx import convert_sklearn
from skl2onnx.common.data_types import FloatTensorType, StringTensorType
from falcon.config import ONNX_OPSET_VERSION, ML_ONNX_OPSET_VERSION
from typing import List, Optional, Type, Any
from sklearn.pipeline import Pipeline as SKLPipeline
from sklearn.preprocessing import MaxAbsScaler, OrdinalEncoder
from skl2onnx.sklapi import CastTransformer
from falcon.serialization import SerializedModelRepr

class ScalerAndEncoder(Processor, ONNXConvertible):
    """
    Applies OneHotEncoder/OrdinalEncoder on low/high cardinality categorical features and StandardScaler on numerical features.
    """
    
    def __init__(
        self, mask: List[int]
    ) -> None:  # [1/2 -> low/high cardinality categorical, 0 -> numerical]
        """
        Parameters
        ----------
        mask : List[int]
            boolean mask with True/False for categorical/numerical features
        """
        self.mask = mask

    def fit(self, X: npt.NDArray, _: Any = None) -> None:
        """
        Fits the encoder.

        Parameters
        ----------
        X : npt.NDArray
            data to encode
        _ : Any, optional
            dummy argument to keep compatibility with pipeline training, by default None
        """
        transformers = []
        for i, v in enumerate(self.mask):
            if v == 1:
                method = OneHotEncoder(
                    categories="auto", sparse=False, handle_unknown="ignore"
                )
            elif v == 0:
                method = SKLPipeline(steps = [('cast64', CastTransformer(dtype=np.float64)),('scaler', StandardScaler(with_mean=True, with_std=True)),('cast32', CastTransformer())])
            else:
                method = SKLPipeline(steps=[('ord_enc', OrdinalEncoder(categories="auto", handle_unknown = "use_encoded_value", unknown_value = -1)), ('sc', MaxAbsScaler())])
            
            t = (f"input {i}", method, [i])
            transformers.append(t)

        self.ct = ColumnTransformer(transformers)
        self.ct.fit(X)

    def predict(self, X: npt.NDArray) -> npt.NDArray:
        """
        Applies the encoder.

        Parameters
        ----------
        X : npt.NDArray
            input data

        Returns
        -------
        npt.NDArray
            encoded data
        """
        return self.ct.transform(X).astype(dtype=np.float32)

    def get_input_type(self) -> Type:
        """
        Returns
        -------
        Type
            npt.NDArray[np.object_]
        """
        return npt.NDArray[np.object_]

    def get_output_type(self) -> Type:
        """
        Returns
        -------
        Type
            Float32Array
        """
        return Float32Array

    def forward(self, X: npt.NDArray[np.object_]) -> npt.NDArray:
        """
        Equivalent of `.predict()` or `.transform()`.


        Parameters
        ----------
        X : npt.NDArray[np.object_]
            data to process

        Returns
        -------
        npt.NDArray
            processed data
        """
        return self.transform(X)

    def to_onnx(self) -> SerializedModelRepr:
        """
        Serializes the encoder to onnx. 
        Each feature in the original dataset is mapped to its own input node (`float32` for numerical or `string` for categorical).

        Returns
        -------
        SerializedModelRepr
        """
        initial_types = []
        initial_types_str: List[str] = []
        initial_shapes: List[List[Optional[int]]] = []
        for i, t in enumerate(self.mask):
            if t == 1 or t == 2:
                tensor = StringTensorType([None, 1])
                initial_types_str.append("STRING")
            else:
                tensor = FloatTensorType([None, 1])
                initial_types_str.append("FLOAT32")
            initial_types.append((f"input{i}", tensor))
            initial_shapes.append([None, 1])
        return SerializedModelRepr(
            convert_sklearn(
                self.ct,
                initial_types=initial_types,
                target_opset={'': ONNX_OPSET_VERSION, 'ai.onnx.ml': ML_ONNX_OPSET_VERSION},
                options={StandardScaler: {"div": "div_cast"}},
            ).SerializeToString(),
            len(self.mask),
            1,
            initial_types_str,
            initial_shapes,
        )



from falcon.abstract.model import TransformerMixin

from falcon.abstract.task_pipeline import PipelineElement


class Processor(PipelineElement, TransformerMixin):
    """
    Subclass of `PipelineElement`. Used for data pre and post processing (e.g. data scaling).
    """
    pass



from typing import Any, Union, Dict, List, Type, Optional
import onnxruntime as ort
import numpy as np
from abc import ABC, abstractmethod
import bson
from bson import BSON


class BaseRuntime(ABC):
    @abstractmethod
    def run(self, X: np.ndarray, **kwargs: Any) -> Any:
        pass


class ONNXRuntime(BaseRuntime):
    """
    Runtime for ONNX models. This runtime can only run onnx models produced by falcon.
    """
    def __init__(self, model: Union[bytes, str]):
        self.ort_session = ort.InferenceSession(model)

    def run(self, X: np.ndarray, outputs: str = "final", **kwargs: Any) -> List[np.ndarray]:
        """
        Runs the model.

        Parameters
        ----------
        X : np.ndarray
            model 
        outputs : str, optional
            when set to "all", all onnx output nodes will be returned; when "final" only the last layer outputs are returned, by default "final"

        Returns
        -------
        List[np.ndarray]
            model predictions
        """
        if outputs not in ["all", "final"]:
            raise ValueError(
                f"Expected `outputs` to be one of [all, final], got `{outputs}`."
            )

        inputs = self._get_inputs(X)
        output_names = self._get_output_names(outputs)

        return self.ort_session.run(output_names, inputs)

    def _get_inputs(self, X: np.ndarray) -> Dict:
        ort_inputs = self.ort_session.get_inputs()
        inputs = {}

        for i, inp in enumerate(ort_inputs):
            dtype: Any
            if str(inp.type) == "tensor(float)":
                dtype = np.float32
            elif str(inp.type) == "tensor(string)":
                dtype = np.str_
            elif "int" in str(inp.type).lower():
                dtype = np.int64
            else:
                RuntimeError(
                    f"The model input type should be one of [str, int, float], got f{inp.type}"
                )
            if len(ort_inputs) > 1:
                inputs[str(inp.name)] = np.expand_dims(X[:, i], 1).astype(dtype)
            else:
                inputs[str(inp.name)] = X.astype(dtype)
        return inputs

    def _get_output_names(self, outputs: str = "final") -> List[str]:
        ort_outputs = self.ort_session.get_outputs()
        output_names = [o.name for o in ort_outputs]
        if outputs == "final" and len(ort_outputs) > 1:
            idx_ = []
            for name in output_names:
                if not name[0:10] == "falcon_pl_":
                    raise RuntimeError("One of the output nodes has an invalid name.")
                idx = int(name.split("/")[0][10:])
                idx_.append(idx)
            max_idx = max(idx_)
            output_names = [
                n for n in output_names if n.startswith(f"falcon_pl_{max_idx}")
            ]
        return output_names


class FalconRuntime(BaseRuntime):
    """
    Runtime for falcon models.
    This runtime is added for future compatibility in case some models will not be onnx convertible. 
    Ideally, it will not have to be used.
    """
    def __init__(self, model: Union[bson.BSON, str, bytes]):
        """
        Parameters
        ----------
        model : Union[bson.BSON, str, bytes]
            Serialized model or filepath
        """

        if isinstance(model, str) and model.endswith(".falcon"):
            with (open(model, "rb")) as f:
                model = f.read()

        self.decoded_model: Dict = bson.BSON.decode(model) # type: ignore

    def run(self, X: np.ndarray, **kwargs: Any) -> Union[List, np.ndarray]:
        """
        Runs the model.

        Parameters
        ----------
        X : np.ndarray
            model inputs
            

        Returns
        -------
        Union[List, np.ndarray]
            model prediction
        """
        for i, node in enumerate(self.decoded_model["nodes"]):
            if node["type"] == "onnx":
                X = self._run_onnx_node(node, X, self._get_n_inputs_node(i + 1))
            else:
                raise RuntimeError(f"Unknown node type {node['type']}")
        if len(X) == 1:
            X = X[0]
        if len(X.shape) > 1:
            X = np.squeeze(X)
        return X

    def _get_n_inputs_node(self, i: int) -> Optional[int]:
        if len(self.decoded_model["nodes"]) > i:
            return self.decoded_model["nodes"][i]["n_inputs"]
        else:
            return None

    def _run_onnx_node(
        self, node: Dict, X: np.ndarray, n_inputs_next: Optional[int] = None
    ) -> np.ndarray:
        ort_sess = ort.InferenceSession(node["model"])
        ort_inputs = ort_sess.get_inputs()
        inputs = {}
        for oi, inp in enumerate(ort_inputs):
            dtype = self._get_np_type(node["initial_types"][oi])
            if len(ort_sess.get_inputs()) > 1:
                inputs[str(inp.name)] = np.expand_dims(X[:, oi], 1).astype(dtype)
            else:
                if len(X.shape) > len(node["initial_shapes"][oi]):
                    X = X.squeeze()
                elif len(X.shape) < len(node["initial_shapes"][oi]):
                    X = np.expand_dims(X, 1)
                inputs[str(inp.name)] = X.astype(dtype)
        outputs = [o.name for o in ort_sess.get_outputs()]
        print("next >> ", n_inputs_next)
        if n_inputs_next is not None:
            outputs = outputs[:n_inputs_next]
        pred = ort_sess.run(outputs, inputs)
        if len(pred) == 1:
            pred = pred[0]
        else:
            print(len(pred))
            print(pred)
            raise NotImplementedError(  # TODO
                "Multiple outputs from node is not yet supported."
            )
        return pred

    def _get_np_type(self, type_: str) -> Type:
        dtype: Type
        if type_ == "FLOAT32":
            dtype = np.float32
        elif type_ == "STRING":
            dtype = np.str_
        elif type_ == "INT64":
            dtype = np.int64
        else:
            RuntimeError("The model has an unsupported input type")
        return dtype



from falcon.abstract import TaskManager
from falcon.tabular import TabularTaskManager
from typing import Any, Optional, Dict, Type, Union
from falcon.abstract import Pipeline, TaskManager
import warnings
import datetime
from falcon.task_configurations import get_task_configuration

def warn(*args: Any, **kwargs: Any) -> None:
    pass


def initialize(
    task: str,
    data: Any,
    pipeline: Optional[Type[Pipeline]] = None,
    pipeline_options: Optional[Dict] = None,
    extra_pipeline_options: Optional[Dict] = None,
    features: Any = None,
    target: Any = None,
    **options: Any
) -> TaskManager:
    """
    Initializes and returns a task manager object for a given task.

    Parameters
    ----------
    task : str
        type of the task
    data : Any
        data to be used for training
    pipeline : Optional[Type[Pipeline]], optional
        class to be used as pipeline, by default None
    pipeline_options : Optional[Dict], optional
        arguments to be passed to the pipeline, by default None.
        These options will overwrite the ones from `default_pipeline_options` attribute
    extra_pipeline_options : Optional[Dict], optional
        arguments to be passed to the pipeline, by default None.
        These options will be passed in addition to the ones from `default_pipeline_options` attribute.
        This argument is ignored if `pipeline_options` is not None
    features : Any, optional
        features to be used for training, by default None
    target : Any, optional
        target to be used for training, by default None

    Returns
    -------
    TaskManager
        Initialized task manager object
    """
    warnings.warn = warn
    if task == "tabular_classification" or task == "tabular_regression":
        manager: TabularTaskManager = TabularTaskManager(
            task=task,
            data=data,
            pipeline=pipeline,
            pipeline_options=pipeline_options,
            extra_pipeline_options=extra_pipeline_options,
            features=features,
            target=target,
            **options
        )
    else:
        raise ValueError("Invalid task")

    return manager


# TODO Allow NONE for the task
def AutoML(
    task: str,
    train_data: Any,
    test_data: Any =  None,
    features: Any = None,
    target: Any = None,
    manager_configuration: Optional[Union[Dict, str]] = None
) -> TaskManager:
    """
    High level API for one line model training and evaluation.
    
    When calling the following steps will be executed:
        1) task manager object will be initialized;
        2) the model will be trained;
        3) performance summary table is printed (if test set is not provided, random split is done);
        4) the model is saved as an onnx file.

    Parameters
    ----------
    task : str
        type of the task, currently supported tasks are [`tabular_classification`, `tabular_regression`]
    train_data : Any
        data to be used for training, for tabular classification and regression this can be: path to .csv or .parquet file, pandas dataframe, numpy array, tuple (X,y)
    test_data : Any, optional
        data to be used for evaluation, for tabular classification and regression this can be: path to .csv or .parquet file, pandas dataframe, numpy array, tuple (X,y)
    features : Any, optional
        features to be used for training, for tabular classification and regression this can be: list of column names or indexes, by default None
    target : Any, optional
        target to be used for training, for tabular classification and regression this can be: column name or index, by default None
    manager_configuration : Unionp[Dict, str], optional
        task manager configuration to be used (can be used to replace pipeline/learner and/or their arguments), by default None

    Returns
    -------
    TaskManager
        Task Manager object for the corresponding task.
    """
    task = task.lower()
    if manager_configuration is None:
        manager_configuration_ = {}
    elif isinstance(manager_configuration, str): 
        manager_configuration_ = get_task_configuration(task=task, configuration_name=manager_configuration)
    manager = initialize(task=task, data=train_data, features=features, target=target, **manager_configuration_)
    make_eval_subset = True if test_data is None else False
    manager.train(pre_eval = False, make_eval_subset = make_eval_subset)
    manager.performance_summary(test_data = test_data)
    print('Saving the model ...')
    ts = datetime.datetime.now().strftime('%Y%m%d.%H%M%S')
    filename = f"falcon_{ts}.onnx"
    manager.save_model(format = 'onnx', filename = filename)
    print(f"The model was saved as `{filename}`")
    return manager



from typing import Callable
from sklearn.ensemble import (
    AdaBoostRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
)
from sklearn.linear_model import ElasticNet, LinearRegression
from sklearn.base import BaseEstimator as SklearnBaseEstimator
from sklearn.svm import NuSVR, SVR
from falcon.abstract import Learner
from falcon.abstract.onnx_convertible import ONNXConvertible
from falcon.types import Float32Array, Int64Array
from typing import Dict, List, Tuple, Callable, Optional, List, Type, Any, Union

from imblearn.over_sampling import RandomOverSampler
from sklearn.discriminant_analysis import (
    LinearDiscriminantAnalysis,
    # QuadraticDiscriminantAnalysis, # DO NOT USE
)
from sklearn.ensemble import (
    AdaBoostClassifier,
    AdaBoostRegressor,
    BaggingClassifier,
    BaggingRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    HistGradientBoostingClassifier,
    HistGradientBoostingRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import ElasticNet, LinearRegression, LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC, SVR, NuSVC, NuSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from sklearn.model_selection import train_test_split
from sklearn.metrics import r2_score, balanced_accuracy_score

from falcon.tabular.models import StackingRegressor, StackingClassifier
from falcon.utils import print_
import pandas as pd

import numpy as np
from numpy import typing as npt
from falcon.serialization import SerializedModelRepr

_SKLEARN_VERBOSE = 0  # for debugging only

_default_estimators: Dict = {
    "tabular_regression": {
        "mini": [
            ("LinearRegression", LinearRegression, {}),
            ("ElasticNet", ElasticNet, {}),
            ("SVR", SVR, {}),
            ("NuSVR", NuSVR, {}),
            ("DecisionTreeRegressor", DecisionTreeRegressor, {}),
            (
                "HistGradientBoostingRegressor",
                HistGradientBoostingRegressor,
                {"min_samples_leaf": 2},
            ),
            ("GradientBoostingRegressor_100", GradientBoostingRegressor, {}),
            ("AdaBoostRegressor_50", AdaBoostRegressor, {}),
            (
                "BaggingRegressor_10",
                BaggingRegressor,
                {
                    "base_estimator": DecisionTreeRegressor(min_samples_split=2),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_100",
                RandomForestRegressor,
                {"min_samples_split": 2, "n_jobs": -1, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "ExtraTreesRegressor_100",
                ExtraTreesRegressor,
                {"min_samples_split": 2, "n_jobs": -1, "verbose": _SKLEARN_VERBOSE},
            ),
            ("SVR_linear", SVR, {"C": 0.025, "kernel": "linear"}),
            (
                "GradientBoostingRegressor_10",
                GradientBoostingRegressor,
                {"n_estimators": 10},
            ),
            (
                "GradientBoostingRegressor_25",
                GradientBoostingRegressor,
                {"n_estimators": 25},
            ),
            (
                "GradientBoostingRegressor_50",
                GradientBoostingRegressor,
                {"n_estimators": 50},
            ),
            (
                "GradientBoostingRegressor_200",
                GradientBoostingRegressor,
                {"n_estimators": 200},
            ),
            ("AdaBoostRegressor_10", AdaBoostRegressor, {"n_estimators": 10}),
            ("AdaBoostRegressor_25", AdaBoostRegressor, {"n_estimators": 25}),
            ("AdaBoostRegressor_100", AdaBoostRegressor, {"n_estimators": 100}),
            (
                "BaggingRegressor_25",
                BaggingRegressor,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=2),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingRegressor_50",
                BaggingRegressor,
                {
                    "n_estimators": 50,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=2),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingRegressor_100",
                BaggingRegressor,
                {
                    "n_estimators": 100,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=2),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_10",
                RandomForestRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_25",
                RandomForestRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_50",
                RandomForestRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_200",
                RandomForestRegressor,
                {
                    "n_estimators": 200,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_10",
                ExtraTreesRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_25",
                ExtraTreesRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_50",
                ExtraTreesRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_200",
                ExtraTreesRegressor,
                {
                    "n_estimators": 200,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
        "mid": [
            ("ElasticNet", ElasticNet, {}),
            (
                "DecisionTreeRegressor",
                DecisionTreeRegressor,
                {"min_samples_split": 0.001},
            ),
            (
                "HistGradientBoostingRegressor_150",
                HistGradientBoostingRegressor,
                {"max_iter": 150},
            ),
            ("HistGradientBoostingRegressor", HistGradientBoostingRegressor, {}),
            (
                "HistGradientBoostingRegressor_50",
                HistGradientBoostingRegressor,
                {"max_iter": 50},
            ),
            (
                "GradientBoostingRegressor_100",
                GradientBoostingRegressor,
                {"min_samples_split": 0.003},
            ),
            ("AdaBoostRegressor_50", AdaBoostRegressor, {}),
            (
                "BaggingRegressor_10",
                BaggingRegressor,
                {
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.001),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_100",
                RandomForestRegressor,
                {"min_samples_split": 0.003, "n_jobs": 4, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "ExtraTreesRegressor_100",
                ExtraTreesRegressor,
                {"min_samples_split": 0.003, "n_jobs": 4, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "GradientBoostingRegressor_10",
                GradientBoostingRegressor,
                {"n_estimators": 10, "min_samples_split": 0.003},
            ),
            (
                "GradientBoostingRegressor_25",
                GradientBoostingRegressor,
                {"n_estimators": 25, "min_samples_split": 0.003},
            ),
            (
                "GradientBoostingRegressor_50",
                GradientBoostingRegressor,
                {"n_estimators": 50, "min_samples_split": 0.003},
            ),
            ("AdaBoostRegressor_25", AdaBoostRegressor, {"n_estimators": 25}),
            ("AdaBoostRegressor_100", AdaBoostRegressor, {"n_estimators": 100}),
            (
                "BaggingRegressor_25",
                BaggingRegressor,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.003),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingRegressor_50",
                BaggingRegressor,
                {
                    "n_estimators": 50,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.003),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingRegressor_100",
                BaggingRegressor,
                {
                    "n_estimators": 100,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.003),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_10",
                RandomForestRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_25",
                RandomForestRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_50",
                RandomForestRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_10",
                ExtraTreesRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_25",
                ExtraTreesRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_50",
                ExtraTreesRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
        "large": [
            ("ElasticNet", ElasticNet, {}),
            ("HistGradientBoostingRegressor", HistGradientBoostingRegressor, {}),
            (
                "HistGradientBoostingRegressor_200",
                HistGradientBoostingRegressor,
                {"max_iter": 200},
            ),
            (
                "BaggingRegressor_10",
                BaggingRegressor,
                {
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.001),
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_100",
                RandomForestRegressor,
                {"min_samples_split": 0.003, "n_jobs": 1, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "ExtraTreesRegressor_100",
                ExtraTreesRegressor,
                {"min_samples_split": 0.003, "n_jobs": 2, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "GradientBoostingRegressor_25",
                GradientBoostingRegressor,
                {"n_estimators": 25},
            ),
            ("AdaBoostRegressor_10", AdaBoostRegressor, {"n_estimators": 10}),
            ("AdaBoostRegressor_25", AdaBoostRegressor, {"n_estimators": 25}),
            (
                "BaggingRegressor_25",
                BaggingRegressor,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeRegressor(min_samples_split=0.001),
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_10",
                RandomForestRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_25",
                RandomForestRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestRegressor_50",
                RandomForestRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_10",
                ExtraTreesRegressor,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_25",
                ExtraTreesRegressor,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesRegressor_50",
                ExtraTreesRegressor,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
    },
    "tabular_classification": {
        "mini": [
            ("LogisticRegression", LogisticRegression, {}),
            ("DecisionTreeClassifier", DecisionTreeClassifier, {}),
            ("SVC", SVC, {}),
            ("NuSVC", NuSVC, {}),
            ("GaussianNB", GaussianNB, {}),
            ("AdaBoostClassifier_50", AdaBoostClassifier, {}),
            ("GradientBoostingClassifier_100", GradientBoostingClassifier, {}),
            (
                "HistGradientBoostingClassifier",
                HistGradientBoostingClassifier,
                {"min_samples_leaf": 2},
            ),
            (
                "RandomForestClassifier_100",
                RandomForestClassifier,
                {"min_samples_split": 0.003, "n_jobs": -1, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "BaggingClassifier_10",
                BaggingClassifier,
                {
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_100",
                ExtraTreesClassifier,
                {"min_samples_split": 0.003, "n_jobs": -1, "verbose": _SKLEARN_VERBOSE},
            ),
            ("SVC_linear", SVC, {"C": 0.025, "kernel": "linear"}),
            ("LinearDiscriminantAnalysis", LinearDiscriminantAnalysis, {}),
            # ("QuadraticDiscriminantAnalysis", QuadraticDiscriminantAnalysis, {}), # this breaks onnx
            ("AdaBoostClassifier_10", AdaBoostClassifier, {"n_estimators": 10}),
            ("AdaBoostClassifier_25", AdaBoostClassifier, {"n_estimators": 25}),
            ("AdaBoostClassifier_100", AdaBoostClassifier, {"n_estimators": 100}),
            (
                "GradientBoostingClassifier_10",
                GradientBoostingClassifier,
                {"n_estimators": 10},
            ),
            (
                "GradientBoostingClassifier_25",
                GradientBoostingClassifier,
                {"n_estimators": 25},
            ),
            (
                "GradientBoostingClassifier_50",
                GradientBoostingClassifier,
                {"n_estimators": 50},
            ),
            (
                "GradientBoostingClassifier_200",
                GradientBoostingClassifier,
                {"n_estimators": 200},
            ),
            (
                "RandomForestClassifier_10",
                RandomForestClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_25",
                RandomForestClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_50",
                RandomForestClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_200",
                RandomForestClassifier,
                {
                    "n_estimators": 200,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_25",
                BaggingClassifier,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_50",
                BaggingClassifier,
                {
                    "n_estimators": 50,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_100",
                BaggingClassifier,
                {
                    "n_estimators": 100,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_10",
                ExtraTreesClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_25",
                ExtraTreesClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_50",
                ExtraTreesClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_200",
                ExtraTreesClassifier,
                {
                    "n_estimators": 200,
                    "min_samples_split": 2,
                    "n_jobs": -1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
        "mid": [
            ("LogisticRegression", LogisticRegression, {"max_iter": 150}),
            (
                "DecisionTreeClassifier",
                DecisionTreeClassifier,
                {"min_samples_split": 0.001},
            ),
            ("GaussianNB", GaussianNB, {}),
            ("LinearDiscriminantAnalysis", LinearDiscriminantAnalysis, {}),
            ("AdaBoostClassifier_50", AdaBoostClassifier, {}),
            (
                "GradientBoostingClassifier_100",
                GradientBoostingClassifier,
                {"min_samples_split": 0.003},
            ),
            (
                "HistGradientBoostingClassifier_50",
                HistGradientBoostingClassifier,
                {"max_iter": 50},
            ),
            ("HistGradientBoostingClassifier", HistGradientBoostingClassifier, {}),
            (
                "HistGradientBoostingClassifier_150",
                HistGradientBoostingClassifier,
                {"max_iter": 150},
            ),
            (
                "RandomForestClassifier_100",
                RandomForestClassifier,
                {"min_samples_split": 0.003, "n_jobs": 4, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "BaggingClassifier_10",
                BaggingClassifier,
                {
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_100",
                ExtraTreesClassifier,
                {"min_samples_split": 0.003, "n_jobs": 4, "verbose": _SKLEARN_VERBOSE},
            ),
            ("AdaBoostClassifier_25", AdaBoostClassifier, {"n_estimators": 25}),
            ("AdaBoostClassifier_100", AdaBoostClassifier, {"n_estimators": 100}),
            (
                "GradientBoostingClassifier_10",
                GradientBoostingClassifier,
                {"n_estimators": 10, "min_samples_split": 0.003},
            ),
            (
                "GradientBoostingClassifier_25",
                GradientBoostingClassifier,
                {"n_estimators": 25, "min_samples_split": 0.003},
            ),
            (
                "GradientBoostingClassifier_50",
                GradientBoostingClassifier,
                {"n_estimators": 50, "min_samples_split": 0.003},
            ),
            (
                "RandomForestClassifier_10",
                RandomForestClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_25",
                RandomForestClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_50",
                RandomForestClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_25",
                BaggingClassifier,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_50",
                BaggingClassifier,
                {
                    "n_estimators": 50,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_100",
                BaggingClassifier,
                {
                    "n_estimators": 100,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_10",
                ExtraTreesClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_25",
                ExtraTreesClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_50",
                ExtraTreesClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 4,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
        "large": [
            ("LogisticRegression", LogisticRegression, {"max_iter": 250}),
            ("GaussianNB", GaussianNB, {}),
            ("LinearDiscriminantAnalysis", LinearDiscriminantAnalysis, {}),
            ("HistGradientBoostingClassifier", HistGradientBoostingClassifier, {}),
            (
                "HistGradientBoostingClassifier_200",
                HistGradientBoostingClassifier,
                {"max_iter": 200},
            ),
            (
                "HistGradientBoostingClassifier_50",
                HistGradientBoostingClassifier,
                {"max_iter": 50},
            ),
            (
                "RandomForestClassifier_100",
                RandomForestClassifier,
                {"min_samples_split": 0.003, "n_jobs": 2, "verbose": _SKLEARN_VERBOSE},
            ),
            (
                "BaggingClassifier_10",
                BaggingClassifier,
                {
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_100",
                ExtraTreesClassifier,
                {"min_samples_split": 0.003, "n_jobs": 2, "verbose": _SKLEARN_VERBOSE},
            ),
            ("AdaBoostClassifier_200", AdaBoostClassifier, {"n_estimators": 200}),
            (
                "RandomForestClassifier_10",
                RandomForestClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_25",
                RandomForestClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "RandomForestClassifier_50",
                RandomForestClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 1,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "BaggingClassifier_25",
                BaggingClassifier,
                {
                    "n_estimators": 25,
                    "base_estimator": DecisionTreeClassifier(min_samples_split=0.001),
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_10",
                ExtraTreesClassifier,
                {
                    "n_estimators": 10,
                    "min_samples_split": 0.003,
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_25",
                ExtraTreesClassifier,
                {
                    "n_estimators": 25,
                    "min_samples_split": 0.003,
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
            (
                "ExtraTreesClassifier_50",
                ExtraTreesClassifier,
                {
                    "n_estimators": 50,
                    "min_samples_split": 0.003,
                    "n_jobs": 2,
                    "verbose": _SKLEARN_VERBOSE,
                },
            ),
        ],
    },
}


class SuperLearner(Learner, ONNXConvertible):
    """
    Tabular learner which employs StackingModel for construction of meta estimator.
    """

    def __init__(
        self,
        task: str,
        base_estimators: Optional[List[Tuple[str, Callable, Dict]]] = None,
        base_score_threshold: Optional[float] = None,
        cv: Any = None,
        filter_estimators: Optional[bool] = None,
    ) -> None:
        """
        Constructs a meta model which is trained on cross-validated predictions of base estimators.

        Parameters
        ----------
        task : str
            `tabular_classification` or `tabular_regression`
        base_estimators : Optional[List[Tuple[str, Callable, Dict]]], optional
            list of base estimators, by default None
        base_score_threshold : Optional[float], optional
            threshold for filtering of the estimators, by default None
        cv : Any, optional
            number of CV folds or CV custom object, by default None
        filter_estimators : Optional[bool], optional
            when True, the perfomance of the estimators pre-estimated on the subset of training, estimators with the performance below the threshold will not be used for meta model construction, by default None
        """

        if task not in ["tabular_classification", "tabular_regression"]:
            raise ValueError(
                f"Invalid task type. Expected `tabular_classification` or `tabular_regression`, found `{task}`."
            )

        self.base_estimators = base_estimators
        self.task = task
        self.base_score_threshold = base_score_threshold
        self.cv = cv
        self.filter_estimators = filter_estimators

    def _split(
        self, X: npt.NDArray, y: npt.NDArray
    ) -> Tuple[Float32Array, Float32Array, Float32Array, Float32Array]:
        X_train: Float32Array
        y_train: Float32Array
        X_val: Float32Array
        y_val: Float32Array

        if self.task == "tabular_classification":
            X_train, X_val, y_train, y_val = train_test_split(X, y, stratify=y)
            X_train_upsampled: Float32Array
            y_train_upsampled: Float32Array
            X_train_upsampled, y_train_upsampled = RandomOverSampler().fit_resample(
                X_train, y_train
            )
            return X_train_upsampled, X_val, y_train_upsampled, y_val
        X_train, X_val, y_train, y_val = train_test_split(X, y, stratify=None)
        return X_train, X_val, y_train, y_val

    def _calculate_base_score(self, y_hat: Float32Array, y: Float32Array) -> float:
        if self.task == "tabular_classification":
            return balanced_accuracy_score(y, y_hat)
        else:
            return (r2_score(y, y_hat) + 1) / 2

    def _set_size_optimized_config(self, X: Float32Array) -> None:

        volume = X.shape[0] * X.shape[1]

        min_threshold = 80_000  # 5_000 samples with 16 features
        mid_threshold = 4_000_000  # 125_000 samples with 32 featrues / 250_000 samples with 16 features

        if volume < min_threshold:
            print_("Using default config for small dataset")
            cv = 10
            base_estimators = _default_estimators[self.task]["mini"]
            filter_estimators = True
        elif volume < mid_threshold:
            print_("Using default config for mid dataset")
            cv = 5
            base_estimators = _default_estimators[self.task]["mid"]
            filter_estimators = True
        else:
            print_("Using default config for large dataset")
            base_estimators = _default_estimators[self.task]["large"]
            cv = 3
            filter_estimators = False

        if self.cv is None:
            self.cv = cv
        if self.base_estimators is None:
            self.base_estimators = base_estimators
        if self.filter_estimators is None:
            self.filter_estimators = filter_estimators

    def _preselect(
        self, X: Float32Array, y: Float32Array
    ) -> List[
        Tuple[str, Callable]
    ]:  # select estimators to be used in the main training loop
        if self.base_estimators is None:
            raise ValueError("expected base_estimators to be a list, found None")
        selected_estimators: List[Tuple[str, Callable]] = []
        if not self.filter_estimators:
            print(
                "\t -> Skipping filtering of base classifiers => all estimators will be used for final model"
            )
            selected_estimators = [
                (estimator[0], estimator[1](**estimator[2]))
                for estimator in self.base_estimators
            ]
            return selected_estimators
        print_(f"\t -> Filtering base classifiers:")
        if self.base_score_threshold is None:
            if self.task == "tabular_classification":
                n_classes = len(np.unique(y, return_counts=False))
                baseline = 1 / n_classes
                self.base_score_threshold = 1.1 * baseline
                print_(
                    f"\t Using {self.base_score_threshold} as baseline score for {n_classes} classes classification task"
                )
            else:
                self.base_score_threshold = 0.55
        X_train: Float32Array
        y_train: Float32Array
        X_val: Float32Array
        y_val: Float32Array
        X_train, X_val, y_train, y_val = self._split(X, y)
        for estimator in self.base_estimators:

            est: SklearnBaseEstimator = estimator[1](**estimator[2])
            est.fit(X_train, y_train)
            y_hat: Float32Array = est.predict(X_val)
            base_score: float = self._calculate_base_score(y_hat, y_val)
            print_(f"\t\t-> {estimator[0]} base score: {base_score}")
            if base_score >= self.base_score_threshold:
                selected_estimators.append((estimator[0], estimator[1](**estimator[2])))
        if len(selected_estimators) < 3:  # using all estimators
            for estimator in self.base_estimators:
                selected_estimators = []
                selected_estimators.append((estimator[0], estimator[1](**estimator[2])))
        return selected_estimators

    def fit(self, X: Float32Array, y: Float32Array) -> None:
        """
        Fits the model. The hyperparameters that were not passed to the `__init__` will be automatically determined based on the size of the training set.
        For classification tasks, the dataset will be balanced by upsampling the minority class(es).

        Parameters
        ----------
        X : Float32Array
            features
        y : Float32Array
            targets
        """
        print_("Fitting stacked model... ")
        self._set_size_optimized_config(X)
        estimators: List[Tuple[str, Callable]] = self._preselect(X, y)
        stacked_estimator: SklearnBaseEstimator
        print_(f"\t -> Fitting the final estimator")
        if self.task == "tabular_classification":
            stacked_estimator = StackingClassifier(
                estimators=estimators, final_estimator=LogisticRegression(), cv=self.cv
            )
        else:
            stacked_estimator = StackingRegressor(
                estimators=estimators, final_estimator=LinearRegression(), cv=self.cv
            )

        stacked_estimator.fit(X, y)
        self.model = stacked_estimator

    def predict(self, X: Float32Array) -> Union[Float32Array, Int64Array]:
        """
        Makes a prediction for given X.

        Parameters
        ----------
        X : Float32Array
            features

        Returns
        -------
        Union[Float32Array, Int64Array]
            predictions
        """
        return self.model.predict(X)

    def get_input_type(self) -> Type:
        """
        Returns
        -------
        Type
            Float32Array
        """
        return Float32Array

    def get_output_type(self) -> Type:
        """
        Returns
        -------
        Type
            Float32Array for regression, Int64Array for classification
        """
        return Float32Array if self.task == "tabular_regression" else Int64Array

    def forward(self, X: Float32Array) -> Float32Array:
        """
        Equivalen to `.predict(X)`

        Parameters
        ----------
        X : Float32Array
            features

        Returns
        -------
        Float32Array
            predictions
        """
        return self.model.predict(X)

    def fit_pipe(self, X: Float32Array, y: Float32Array) -> None:
        """
        Equivalent to `.fit(X, y)`

        Parameters
        ----------
        X : Float32Array
            features
        y : Float32Array
            targets
        """
        self.fit(X, y)

    def to_onnx(self) -> SerializedModelRepr:
        """
        Serializes the underlying model to onnx by calling its `.to_onnx()` method.

        Returns
        -------
        SerializedModelRepr
        """
        return self.model.to_onnx()



__version__ = '0.4.0'
__author__ = 'Oleg Kostromin, Marco Pasini, Iryna Kondrashchenko'

from falcon.main import initialize, AutoML
from falcon.utils import disable_warnings, run_model

disable_warnings()